if ($LaunchMode !$= "normal")
	return;

// Delete existing DSO files...
if (isfile("scripts/ChatGui.cs.dso"))
	deleteFile("scripts/ChatGui.cs.dso");
if (isfile("gui/ChannelBanDlg.gui.dso"))
	deleteFile("gui/ChannelBanDlg.gui.dso");
if (isfile("gui/ChannelKeyDlg.gui.dso"))
	deleteFile("gui/ChannelKeyDlg.gui.dso");
if (isfile("gui/ChannelOptionsDlg.gui.dso"))
	deleteFile("gui/ChannelOptionsDlg.gui.dso");
if (isfile("gui/ChatGui.gui.dso"))
	deleteFile("gui/ChatGui.gui.dso");
if (isfile("gui/ChatOptionsDlg.gui.dso"))
	deleteFile("gui/ChatOptionsDlg.gui.dso");


$IRC::Flag::Ignore = 1;
$IRC::Flag::Away = 2;
$IRC::Flag::Voice = 4;
$IRC::Flag::HalfOp = 8;
$IRC::Flag::Operator = 16;

$IRC::Char["~"] = $IRC::Flag::Operator;	$IRC::Mode["q"] = $IRC::Flag::Operator;
$IRC::Char["@"] = $IRC::Flag::Operator;	$IRC::Mode["o"] = $IRC::Flag::Operator;
$IRC::Char["&"] = $IRC::Flag::Operator;	$IRC::Mode["a"] = $IRC::Flag::Operator;
$IRC::Char["%"] = $IRC::Flag::HalfOp;	$IRC::Mode["h"] = $IRC::Flag::HalfOp;
$IRC::Char["+"] = $IRC::Flag::Voice;	$IRC::Mode["v"] = $IRC::Flag::Voice;

$IRC::CFlag::Private = 1;	$IRC::CMode["p"] = $IRC::CFlag::Private;
$IRC::CFlag::Moderated = 2;	$IRC::CMode["m"] = $IRC::CFlag::Moderated;
$IRC::CFlag::Invite = 4;	$IRC::CMode["i"] = $IRC::CFlag::Invite;
$IRC::CFlag::Limited = 8;	$IRC::CMode["l"] = $IRC::CFlag::Limited;
$IRC::CFlag::NoExtern = 16;	$IRC::CMode["n"] = $IRC::CFlag::NoExtern;
$IRC::CFlag::Secret = 32;	$IRC::CMode["s"] = $IRC::CFlag::Secret;
$IRC::CFlag::TopicLim = 64;	$IRC::CMode["t"] = $IRC::CFlag::TopicLim;
$IRC::CFlag::HasKey = 128;	$IRC::CMode["k"] = $IRC::CFlag::HasKey;


$IRC::Timeout::CTCP = 5;


////////////////////////////////////////////////////////////////////////////////


if (!isObject(TNIRC))
{
	new ScriptObject(TNIRC)
	{
		class = "IRC";

		nick = getField(wonGetAuthInfo(),0) @ "|T2";
		ident = getField(wonGetAuthInfo(),0);

		maxRetries = 4;
		retryCount = 0;
	};
	TNIRC.serverPane = new ChannelVector();
}


////////////////////////////////////////////////////////////////////////////////


function IRC::Connect(%this, %server, %port, %password)
{
	if (%this.isConnected())
		return;

	if (%this.ident $= "" || %this.nick $= "")
	{
		error("IRC::connect(): You must set .ident and .nick before connecting!");
		return;
	}

	// Ensure a nice IRC friendly nick!
	%this.nick = IRC_CleanNick(%this.nick);

	if (!%port)
		%port = 6667;

	%this.socket = new TCPObject(IRC_TCP) {owner = %this;};
	if (!isobject(%this.channels))
		%this.channels = new SimGroup() {owner = %this;};
	if (!isobject(%this.users))
		%this.users = new SimGroup(){owner = %this;};

	// Store for re-connection purposes
	%this.pass = %pass;
	%this.server = %server;
	%this.port = %port;

	// Connect
	%this.socket.connect("IP:" @ %server @ ":" @ %port);
}


function IRC::Reconnect(%this)
{
	if (%this.isConnected())
		return;

	%this.reconnecting = true;

	%this.retryCount++;
	if (%this.retryCount <= %this.maxRetries)
		%this.connect(%this.server, %this.port, %this.pass);
}


function IRC::Disconnect(%this, %mesg)
{
	if (!%this.isConnected())
		return;

	if (%mesg $= "")
		%mesg = $pref::IRC::QuitMessage;

	%this.send("QUIT :" @ %mesg);
	%this.onQuit(%this.nick @ "!", %mesg);
	%this.socket.disconnect();
	%this.socket.onDisconnect();
}


function IRC::isConnected(%this)
{
	return isObject(%this.socket);
}


////////////////////////////////////////////////////////////////////////////////


function IRC::onDNSFailed(%this)
{
	messageBoxOK("ALERT", "An error occured while connecting to the IRC server.\n[DNS failure]");
}


function IRC::onConnectFailed(%this)
{
	switch(%this.retryCount)
	{
	case 0:
		messageBoxYesNo("ALERT", "Connection attempt to the IRC server failed, retry?", "TNIRC.reconnect();", "Canvas.popDialog(JoinChatDlg);");
	case %this.maxRetries:
		messageBoxOK("ERROR", "Unable to establish a connection with the IRC server.  Check your settings, or perhaps it's down?");
	default:
		%this.reconnect();
	}
}


function IRC::onTimeout(%this)
{
	%this.onDisconnect();
	messageBoxYesNo("ALERT", "Connection to the IRC server timed-out, reconnect?", "TNIRC.reconnect();", "");
}


function IRC::onDisconnect(%this)
{
	%this.connOK = false;
	%this.gotMOTD = false;
	ChatOpt_Conn.setActive(true);
	ChatOpt_DConn.setActive(false);
}


function IRC::onConnected(%this)
{
	%this.retryCount = 0;

	if (%this.pass !$= "")
		%this.send("PASS " @ %this.pass);

	%this.send("NICK " @ %this.nick);
	%this.send("USER T2IRC 8 * :" @ %this.ident);

	ChatOpt_Conn.setActive(false);
	ChatOpt_DConn.setActive(true);
	ChatOpenPaneBtn.setActive(true);
}


////////////////////////////////////////////////////////////////////////////////


function IRC::onLine(%this, %raw)
{
	if (%this.debug)
		error(%raw);

	//:source EVENT/PARAMS :message
	if (getsubstr(%raw, 0,1) $= ":")
	{
		//:siden.irc.arloria.net 001 torquy :Welcome to the ArloriaNET IRC 
		%line = getsubstr(%raw, 1, 999);
		%line = nextToken(%line, "source", " ");
		%data = nextToken(%line, "command", " ");
	}
	//COMMAND :string (only PING?)
	else
		%data = nextToken(%raw, "command", ":");

	%data = trim(%data);
	%source = trim(%source);
	nextToken(%source, "nick", "!");
	if ((%pos = strpos(%data, ":")) != -1 && getsubstr(%data, %pos+1, 1) $= ":")
		%data = getsubstr(%data, 0, %pos + 1) SPC getsubstr(%data, %pos + 1, 999);

	switch$(trim(%command))
	{
	//
	// TEXTUAL
	//
	case "PING":// PING :string to return
		%this.send("PONG :" @ %data);
	case "PRIVMSG":// :nick!ident@host PRIVMSG #channel|nick :message
		%data = trim(nextToken(%data, "target", ":"));
		%target = trim(%target);

		if (!%this.findUser(%nick).ignore)
		{
			if (getsubstr(%data, 0,1) $= "\x01")
				%this.onCTCP(%nick, %target, %data);
			else
				%this.onMessage(%nick, %target, %data);
		}
	case "NOTICE":// :nick!ident@host NOTICE nick :message
		%data = trim(nextToken(%data, "target", ":"));
		%target = trim(%target);

		if (!%this.findUser(%nick).ignore)
			%this.onNotice(%source, %data);
	case "MODE":// :nick!ident@host MODE #channel +/-modes targets
				// :nick MODE nick :+/-modes
		if (strpos(%data, ":") != -1)
		{
			%this.onServerMesg("Mode change [" @ nextToken(%data, "nil", ":") @ "] for user " @ %this.nick);
			return;
		}

		%data = nextToken(%data, "channel", " ");
		%users = nextToken(%data, "modes", " ");
		%this.show(%channel, "\c9*" SPC %nick SPC "sets mode" SPC %channel SPC %data);
		%this.processModes(%channel, %users, %modes);
		if (%this.currentChannel == %this.channel[%channel])
			%this.refreshUserList();
	case "KICK":// :nick!ident@host KICK #channel target :message
		%data = nextToken(%data, "target", ":");
		%target = nextToken(%target, "chan", " ");
		%this.onKick(%source, trim(%chan), trim(%target), trim(%data));
	case "JOIN":// :nick!ident@host JOIN :#channel
		%data = getsubstr(%data, 1, 999);
		%this.onJoin(%data, %source);
	case "PART":// :nick!ident@host PART #channel :message
		%data = nextToken(%data, "target", ":");
		%this.onPart(trim(%target), %source, trim(%data));
	case "NICK":// :nick!ident@host NICK :new-nick
		%data = getsubstr(%data, 1, 999);
		%this.onNick(%source, %data);
	case "QUIT":// :nick!ident@host QUIT :message
		%data = getsubstr(%data, 1, 999);
		%this.onQuit(%source, %data);
	case "TOPIC":// :nick!ident@host TOPIC #channel :topic
		%data = nextToken(%data, "target", ":");
		rubyEval("tsEval \"$temp=\\\"#{Time.now.to_i.to_s}\\\";\"");
		%this.onTopic(trim(%target), trim(%data), %nick, $temp);
	case "INVITE":// :nick!ident@host INVITE self :#channel
		%channel = nextToken(%data, "data", ":");
		%message = "\c9*" SPC %nick SPC "has invited you to <t2server:" @
			%channel @ ">" @ %channel @ "</t2server>";
		%this.show(%this.currentChannel.getName(), %message);

	//
	// NUMERICS
	//
	case "004":// :server 004 self some-information
		%this.connOK = 1;

		if (!%this.reconnecting)
		{
			// Auto-join channels on connection...
			%str = $pref::IRC::Channels;
			while (%str !$= "")
			{
				%str = nextToken(%str, "chan", " ");
				%this.send("JOIN " @ %chan);
			}
		}
		// Rejoin open channels on reconnection
		else
		{
			for (%i = 0; %i < %this.channels.getCount(); %i++)
				if (!(%chan = %this.channels.getObject(%i)).private)
					%this.send("JOIN" SPC %chan.getName() SPC %chan.key);
		}

	// Channel listing
	case "322":// :server 322 self #channel visible :topic
		%data = nextToken(%data, "params", ":");
		%channel = getWord(%params, 1);
		%users = getWord(%params, 2);

		if (trim(stripchars(%channel, "#")) !$= "" && %users > 0)
		{
			%this.channelInfo[%this.channelInfoCt] = getWord(%params, 1) TAB getWord(%params, 2);
			%this.channelInfoCt++;
		}
	case "323":// :server 323 self :End of /list
		%this.onChannelListUpdated();

	// Channel topic
	case "332":// :server 332 self #channel :topic
		%data = nextToken(%data, "params", ":");
		%this.tempTopic[getWord(%params, 1)] = trim(%data);
	case "333":// :server 333 self #channel set-by timestamp
		%channel = getWord(%data, 1);
		%nick = getWord(%data, 2);
		%time = getWord(%data, 3);
		%this.onTopic(%channel, %this.tempTopic[%channel], %nick, %time);
		%this.tempTopic[%channel] = "";

	// User list
	case "353":// :server 353 self = #channel :user list
		%channel = getWord(%data, 2);
		%data = nextToken(%data, "params", ":");
		while (%data !$= "")
		{
			%data = nextToken(%data, "user", " ");
			%this.addUser(%channel, %user);
		}
	case "352":// :server 352 self #channel ident host server nick H|G[modes] :hops realname away-message
		%chan = getWord(%data, 1);
		%nick = getWord(%data, 5);
		%login = %nick @ "!" @ getWord(%data, 2) @ "@" @ getWord(%data, 3);
		%status = getWord(%data, 6);

		if (%chan !$= "" && (%user = %this.findUser(%nick)) != -1)
		{
			%user.login = %login;
			for (%i = 1; %i < strlen(%status); %i++)
			{
				%char = getsubstr(%status, %i, 1);
				if ((%mode = $IRC::Char[%char]))
					%modes |= %mode;
			}
			if (getsubstr(%status, 0,1) $= "G")
				%modes |= $IRC::Flag::Away;
			%user.mode[%chan] = %modes;
		}
	case "315":// :server 315 self #channel :End of /WHO list.
		%this.onUserlistUpdated(getWord(%data,1));

	// Nick changing
	case "432":// :server 432 self nick :Erroneous Nickname: Illegal characters
		// They're stupid, and typed something wrong...
		if (%this.connOK)
			%this.onErrorMesg(nextToken(%data, "nil", ":"));
		// the fuck?
		else
		{
			%this.nick = CleanIrcNick(%this.nick @ "-" @ getRandom(0,99));

			if (%this.pass !$= "")
				%this.send("PASS " @ %this.pass);
			%this.send("NICK " @ %this.nick);
			%this.send("USER T2IRC 8 * :" @ %this.ident);
		}
	case "433":// :server 433 self nick :Nickname is already in use.
		// They tried using somebody elses...
		if (%this.connOK)
			%this.onErrorMesg(nextToken(%data, "nil", ":"));
		// Somebody else is connected, find another
		else
		{
			%this.nick = CleanIrcNick(%this.nick @ "-" @ getRandom(0,99));

			if (%this.pass !$= "")
				%this.send("PASS " @ %this.pass);
			%this.send("NICK " @ %this.nick);
			%this.send("USER T2IRC 8 * :" @ %this.ident);
		}

	// Ban List
	case "367":// :server 367 self #channel mask set-by timestamp
		error("BAN]" SPC %data);
		%this.banInfo[%this.banInfoCt] = getWord(%data, 2);
		%this.banInfoCt++;
	case "368":// :server 368 self #channel :End of Channel Ban List
		%this.onBanListUpdated();

	case "475":// :server self #channel :Cannot join channel (+k)
		%this.keyChannel = getWord(%data, 1);
		KeyChannelName.setValue(%this.keyChannel);
		Canvas.pushDialog(ChannelKeyDlg);

	// MOTD
	case "376":
		%this.gotMOTD = 1;
	case "422":
		%this.gotMOTD = 1;


	//
	// Capture these to suppress unnecessary output...
	//
	case "321":// :server 321 self Channel :Users  Name
	case "366":// :server 366 self #channel :End of /NAMES list.

	default:
		// Likely something they want/need to see
		if (%this.connOK && %this.gotMOTD)
			%this.onErrorMesg(restWords(%data));
		// probably part of the MOTD
		else
			%this.onServerMesg(restWords(%data));
	}
}


////////////////////////////////////////////////////////////////////////////////


function IRC::onAction(%this, %nick, %target, %action)
{
	if ($pref::IRC::Highlight)
		%action = %this.highlight(%action, "\c8");

	// Assume it's a private chat...
	if (getsubstr(%target, 0,1) !$= "#")
		%target = %nick;

	%this.show(%target, "\c8*" SPC %nick SPC %action);
}


function IRC::onMessage(%this, %nick, %target, %message)
{
	if ($pref::IRC::Highlight)
		%message = %this.highlight(%message, "\c0");

	// Assume it's a private chat...
	if (%target $= %this.nick)
		%target = %nick;

	%this.show(%target, %nick @ ":" SPC %message);
}


function IRC::onNotice(%this, %source, %message)
{
	%this.onServerMesg(%message);
}


function IRC::onCTCP(%this, %nick, %target, %message)
{
	%message = stripChars(%message, "\x01");
	%args = nextToken(%message, "command", " ");

	if (%command !$= "ACTION")
	{
		// Only accept CTCP from the same person every X seconds, except ACTIONs
		if (%this.lastCTCP[%nick] > getSimTime() + (1000 * $IRC::Timeout::CTCP))
			return;

		%this.lastCTCP[%nick] = getSimTime();
	}

	switch$(%command)
	{
	case "VERSION":
		%str = "TN-IRC v1.0 Tribes2" SPC getT2VersionNumber();
		%this.send("NOTICE" SPC %nick SPC "\x01VERSION" SPC %str @ "\x01");
	case "PING":
		%this.send("NOTICE" SPC %nick SPC "\x01PING" SPC %args @ "\x01");
	case "ACTION":
		%this.onAction(%nick, %target, %args);
	}

	if (%command !$= "ACTION")
	{
		%message = "\c1* Received CTCP ("@ %command @") from" SPC %nick;
		if (isObject(%this.currentChannel))
			%this.show(%this.currentChannel.getName(), %message);
		else
			%this.onServerMessage(%message);
	}
}


function IRC::highlight(%this, %message, %color)
{
	%wc = getWordCount(%message);
	for (%i = 0; %i < %wc; %i++)
	{
		%w = getWord(%message, %i);
		if (%this.nick $= %w)
			%str = %str SPC "\c2" @ %w @ %color;
		else
			%str = %str SPC %w;
	}

	return trim(%str);
}


////////////////////////////////////////////////////////////////////////////////


function IRC::requestBanList(%this)
{
	%this.banInfoCt = 0;
	%this.send("MODE" SPC %this.currentChannel.getName() SPC "+b");
}


function IRC::onBanListUpdated(%this)
{
	ChannelBanList.clear();

	for (%i = 0; %i < %this.banInfoCt; %i++)
		ChannelBanList.addRow(%i, %this.banInfo[%i]);

	ChannelBanList.sort(0);
}


function IRC::RemoveBan(%this)
{
	%idx = ChannelBanList.getSelectedId();
	%this.send("MODE" SPC %this.currentChannel.getName() SPC "-b" SPC %this.banInfo[%idx]);
	%this.requestBanList();
}


////////////////////////////////////////////////////////////////////////////////


function IRC::requestChannelList(%this)
{
	%this.channelInfoCt = 0;
	%this.send("LIST");
}


function IRC::onChannelListUpdated(%this)
{
	JoinChatList.clear();
	for (%i = 0; %i < %this.channelInfoCt; %i++)
	{
		%name = getField(%this.channelInfo[%i], 0);
		%temp = (%name $= "#the_construct" || strpos(strlwr(%name), "tribes") != -1);

		JoinChatList.addRow(%i, %this.channelInfo[%i] TAB %temp);
		JoinChatList.setRowStyle(%i, %temp > 0);
	}

	JoinChatList.sort();
	JoinChatName.onCharInput();
}


////////////////////////////////////////////////////////////////////////////////


function IRC::onUserListUpdated(%this, %channel)
{
	if (%this.currentChannel == %this.channel[%channel])
		%this.refreshUserList();

	cancel(%this.who[%channel]);
	%this.who[%channel] = %this.schedule(30000, "send", "WHO" SPC %channel);
}


function IRC::refreshUserList(%this)
{
	%chan = %this.currentChannel;
	if (!isObject(%chan))
		return;

	%channel = %chan.getName();

	ChatRoomMemberList.clear();
	for (%i = 0; %i < %chan.users; %i++)
	{
		%user = %chan.user[%i];
		ChatRoomMemberList.addRow(%i, %user.nick);
		ChatRoomMemberList.setRowStyle(%i, %user.mode[%channel]|(%user.ignore ? $IRC::Flag::Ignore : 0));
	}

	ChatEditChannelBtn.setVisible(%this.self().mode[%channel] & ($IRC::Flag::Operator|$IRC::Flag::HalfOp));
}


////////////////////////////////////////////////////////////////////////////////


function IRC::onTopic(%this, %channel, %topic, %setBy, %time)
{
	%this.channel[%channel].topic = %topic;

	%this.show(%channel, "\c9* Topic for " @ %channel @ " is: " @ %topic);
	%this.show(%channel, "\c9* Set by " @ %setBy @ " on " @ IRC_TimeString(%time));

	if (%this.currentChannel == %this.channel[%channel])
		ChatChannelTopic.setValue(%topic);
}


function IRC::processModes(%this, %channel, %users, %modes)
{
	%idx = 0;
	%len = strlen(%modes);

	for (%i = 0; %i < %len; %i++)
	{
		%char = getsubstr(%modes, %i, 1);
		switch$(%char)
		{
		case "+":
			%give = true;
		case "-":
			%give = false;
		default:
			if ((%mode = $IRC::Mode[%char]))
			{
				%user = %this.findUser(getWord(%users, %idx));
				%idx++;
				if (%give)
					%user.mode[%channel] |= %mode;
				else
					%user.mode[%channel] &= ~%mode;
			}
			else if ((%mode = $IRC::CMode[%char]))
			{
				if (%give)
					%this.channel[%channel].mode |= %mode;
				else
					%this.channel[%channel].mode &= ~%mode;
			
				if (%char $= "k")
				{
					%this.channel[%channel].key = getWord(%users, %idx);
					%idx++;
				}
				if (%char $= "l")
				{
					%this.channel[%channel].limit = getWord(%users, %idx);
					%idx++;
				}
			}
		}
	}
}


function IRC::onJoin(%this, %channel, %user)
{
	nextToken(%user, "nick", "!");

	// Guess we just joined
	if (%this.nick $= %nick)
	{
		%this.addChannel(%channel);
		%this.show(%channel, "\c3You have joined" SPC %channel);
	}
	else
		%this.show(%channel, "\c3" @ %nick SPC "(" @ %user @ ") has joined" SPC %channel);

	%this.addUser(%channel, %nick, %user);
	%this.schedule(1000, "send", "WHO" SPC %channel);

	if (%this.currentChannel.getName() $= %channel)
		%this.refreshUserList();
}


function IRC::onPart(%this, %channel, %user, %message)
{
	nextToken(%user, "nick", "!");

	if (%this.nick $= %nick)
		%this.removeChannel(%channel);
	else
	{
		%this.removeUser(%channel, %nick);
		%this.show(%channel, "\c3" @ %nick SPC "(" @ %user @ ") has left" SPC %channel SPC "(" @ %message @ ")");

		if (%this.currentChannel.getName() $= %channel)
			%this.refreshUserList();
	}
}


function IRC::onKick(%this, %kicker, %channel, %kickee, %reason)
{
	%this.removeUser(%channel, %kickee);
	nextToken(%kicker, "nick", "!");

	if (%kickee $= %this.nick)
		%message = "You have been kicked from "@ %channel @ " by " @ %nick;
	else
		%message = %nick @ " has kicked " @ %kickee @ " from " @ %channel;

	%this.show(%channel, "\c3" @ %message SPC "(" @ %reason @ ")");
	if (%this.currentChannel.getName() $= %channel)
		%this.refreshUserList();
}


////////////////////////////////////////////////////////////////////////////////


function IRC::onNick(%this, %user, %new)
{
	nextToken(%user, "nick", "!");

	if (%this.nick $= %nick)
	{
		%this.nick = %new;
		%message = "\c3You are now known as" SPC %new;
	}
	else
		%message = "\c3" @ %nick SPC "is now known as" SPC %new;

	// Update all channels...
	for (%i = 0; %i < %this.channels.getCount(); %i++)
	{
		%chan = %this.channels.getObject(%i);
		if (%chan.hasUser(%nick))
		{
			%this.show(%chan.getName(), %message);
			if (%chan.private)
				%this.updateChannel(%chan, %new);
		}
	}

	%this.findUser(%nick).nick = %new;
	%this.refreshUserList();
}


function IRC::onQuit(%this, %user, %message)
{
	nextToken(%user, "nick", "!");

	if (%this.nick $= %nick)
	{
		%message = "\c3Quit (" @ %message @ ")";
		for (%i = 0; %i < %this.channels.getCount(); %i++)
			%this.show(%this.channels.getObject(%i).getName(), %message);
	}
	else
	{
		%message = "\c3" @ %nick SPC "(" @ %user @ ")" SPC "has quit (" @ %message @ ")";
		for (%i = 0; %i < %this.channels.getCount(); %i++)
		{
			%chan = %this.channels.getObject(%i);
			if (%this.removeUser(%chan.getName(), %nick))
			{
				%this.show(%chan.getName(), %message);
				if (%chan == %this.currentChannel)
					%this.refreshUserList();
			}
		}
	}
}


////////////////////////////////////////////////////////////////////////////////


function IRC::Show(%this, %target, %message)
{
	if (!isobject(%this.channel[%target]))
		%this.addChannel(%target);

	%channel = %this.channel[%target];
	%channel.pushBackLine(%message);
	if (%channel.getNumLines() > 500)
		%channel.popFrontLine();
}


function IRC::onServerMesg(%this, %message)
{
	%this.serverPane.pushBackLine(%message);
	if (%this.serverPane.getNumLines() > 500)
		%this.serverPane.popFrontLine();
}


function IRC::onErrorMesg(%this, %message)
{
	%message = "\c8" @ %message;

	if (isobject(%this.currentChannel))
		%this.show(%this.currentChannel.getName(), %message);
	else
		%this.onServerMesg(%message);
}


function IRC::Send(%this, %data)
{
	if (%this.isConnected())
		%this.socket.send(%data @ "\r\n");
}


function IRC::SendMessage(%this, %target, %message)
{
	if (getsubstr(%message, 0,1) $= "/")
	{
		%message = getsubstr(%message, 1, 999);

		if (getsubstr(%message, 0,1) !$= "/")
		{
			%this.runCommand(firstWord(%message), restWords(%message), %target);
			return;
		}
	}

	%this.onMessage(%this.nick, %target, %message);
	%this.send("PRIVMSG" SPC %target SPC ":" @ %message);
}


function IRC::runCommand(%this, %command, %params, %target)
{
	switch$(%command)
	{
	case "CLEAR":
		%this.currentChannel.clear();
		return;
	case "J":
		%command = "JOIN";
	case "ME" or "ACTION":
		%this.runCommand("CTCP", %target SPC "ACTION" SPC %params);
		%this.onAction(%this.nick, %target, %params);
		return;
	case "CTCP":
		%this.send("PRIVMSG" SPC firstWord(%params) SPC "\x01" @ strupr(getWord(%params,1)) SPC getWords(%params,2) @ "\x01");
		return;
	case "MSG" or "PRIVMSG":
		%command = "PRIVMSG";
		%params = firstWord(%params) SPC ":" @ restWords(%params);
	case "RAW":
		%this.send(%params);
		return;
	case "QUIT":
		%this.onQuit(%this.nick, %params);
		%this.send("QUIT :" @ %params);
		%this.socket.disconnect();
		%this.schedule(0, "onDisconnect");
	}

	if (%command $= "PRIVMSG")
		%this.onLine(":" @ %this.nick SPC "PRIVMSG" SPC %params);

	%this.send(strupr(%command) SPC %params);
}


////////////////////////////////////////////////////////////////////////////////


function IRC::AddChannel(%this, %channel)
{
	// Already exists, assume we've rejoined after a disconnect
	if (%this.channel[%channel] !$= "")
	{
		%vec = %this.channel[%channel];
		for (%i = 0; %i < %vec.users; %i++)
			%vec.user[%i].decrement();
		%vec.users = 0;
	}
	// Create a new one!
	else
	{
		%vec = new ChannelVector(%channel){users = 0;};
		%this.channels.add(%vec);
		%this.channel[%channel] = %vec;
		%vec.private = getsubstr(%channel, 0,1) !$= "#";
		if (%vec.private)
			%this.addUser(%channel, %channel);
		ChatTabView.addTab(%vec, %channel, %vec.private);
		ChatTabView.setSelected(%vec);
	}
	return %vec;
}


function IRC::updateChannel(%this, %chan, %new)
{
	%this.channel[%chan.getName()] = "";
	%chan.setName(%new);
	%this.channel[%new] = %chan;
	ChatTabView.setTabText(%chan, %new);
}


function IRC::RemoveChannel(%this, %channel)
{
	if (%this.channel[%channel] $= "")
		return;

	%chan = %this.channel[%channel];
	for (%i = 0; %i < %chan.users; %i++)
		%chan.user[%i].decrement();

	cancel(%this.upd[%channel]);
	%this.upd[%channel] = "";
	%this.channel[%channel] = "";
	ChatTabView.removeTab(%chan);
	%chan.delete();
}


////////////////////////////////////////////////////////////////////////////////


function IRC::AddUser(%this, %channel, %nick, %login)
{
	%channel = %this.channel[%channel];

	// Create a new User, if doesn't exist
	if ((%user = %this.findUser(%nick)) == -1)
	{
		%user = new ScriptObject() {
			class = "IRC_USER";
			login = %login;
			refCount = 0;
		};
		%this.users.add(%user);
	}
	// Don't allow duplicates...
	for (%i = 0; %i < %channel.users; %i++)
		if (%channel.user[%i].nick $= %user.nick)
			return;

	// Strip "mode" chars from their nick
	for (%i = 0; %i < strlen(%nick); %i++)
	{
		%char = getsubstr(%nick, %i, 1);

		if ((%mode = $IRC::Char[%char]))
			%modes |= %mode;
		else
			%name = %name @ %char;
	}

	%user.nick = %name;
	%user.mode[%channel.getName()] = %modes;

	%channel.user[%channel.users] = %user;
	%channel.users++;
	%user.increment();
}


function IRC::RemoveUser(%this, %channel, %nick)
{
	%channel = %this.channel[%channel];

	for (%idx = 0; %idx < %channel.users; %idx++)
		if (%channel.user[%idx].nick $= %nick)
			break;

	if (%idx < %channel.users)
	{
		%channel.user[%idx].decrement();

		for (%i = %idx; %i < %channel.users; %i++)
			%channel.user[%i] = %channel.user[%i + 1];

		%channel.users--;
		return true;
	}

	return false;
}


function IRC::FindUser(%this, %nick)
{
	%nick = IRC_CleanNick(%nick);

	for (%i = 0; %i < %this.users.getCount(); %i++)
		if ((%user = %this.users.getObject(%i)).nick $= %nick)
			return %user;

	return -1;
}


function IRC::Self(%this)
{
	if (isobject(%this.users))
		return %this.users.getObject(0);
}


////////////////////////////////////////////////////////////////////////////////


function IRC_USER::Increment(%this)
{
	%this.refCount++;
}


function IRC_USER::Decrement(%this)
{
	%this.refCount--;
	if (%this.refCount <= 0)
		%this.delete();
}


function ChannelVector::HasUser(%this, %nick)
{
	for (%i = 0; %i < %this.users; %i++)
		if (%this.user[%i].nick $= %nick)
			return true;

	return false;
}


////////////////////////////////////////////////////////////////////////////////


function IRC_TCP::onConnected(%this)
{
	%this.owner.onConnected();
}


function IRC_TCP::onDNSFailed(%this)
{
	%this.owner.onDNSFailed();
}


function IRC_TCP::onConnectFailed(%this)
{
	%this.owner.onConnectFailed();
}


function IRC_TCP::onConnected(%this)
{
	%this.owner.onConnected();
}


function IRC_TCP::onLine(%this, %line)
{
	cancel(%this.timeoutThread);
	%this.timeoutThread = %this.schedule(60000, "timeout");
	%this.owner.onLine(%line);
}


function IRC_TCP::onDisconnect(%this)
{
	cancel(%this.timeoutThread);
	%this.owner.onDisconnect();
	%this.delete();
}


function IRC_TCP::timeout(%this)
{
	%this.owner.onTimeout();
	%this.disconnect();
	%this.delete();
}


////////////////////////////////////////////////////////////////////////////////


function IRC_TimeString(%ts)
{
	rubyEval("tsEval \"$temp=\\\"#{Time.at(" @ %ts @ ").strftime('%a %b %d, %Y @ %H:%M:%S')}\\\";\"");
	return $temp;
}


function IRC_CleanNick(%nick)
{
	%alpha = "_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
	%numeric = "0123456789-";
	%special = "[]\\`^{}|";

	%clean = "";
	%nick = strreplace(%nick, " ", "_");
	%len = strlen(%nick);

	%char = getsubstr(%nick, 0,1);
	if (strpos(%alpha @ %special, %char) != -1)
		%clean = %char;

	for (%i = 1; %i < %len; %i++)
	{
		%char = getsubstr(%nick, %i,1);
		if (strpos(%alpha, %char) != -1)
			%clean = %clean @ %char;
		else if (strpos(%numeric, %char) != -1)
			%clean = %clean @ %char;
		else if (strpos(%special, %char) != -1)
			%clean = %clean @ %char;
	}

	return %clean;
}

